@@ -3,6 +3,7 @@ import re
 import shutil
 import logging
 from binascii import b2a_hex
+from time import time
 from datetime import datetime
 from cStringIO import StringIO
 from ConfigParser import ConfigParser
@@ -14,7 +15,8 @@ import tg
 from pylons import app_globals as g
 os.environ['HGRCPATH'] = '' # disable loading .hgrc
 from mercurial import ui, hg, cmdutil, commands
-from mercurial.error import RepoLookupError
+from mercurial.error import RepoLookupError, RevlogError
+from paste.deploy.converters import asint
 
 from pymongo.errors import DuplicateKeyError
 
@@ -24,7 +26,12 @@ from ming.utils import LazyProperty
 
 from allura import model as M
 from allura.lib import helpers as h
-from allura.model.repository import topological_sort, GitLikeTree, zipdir
+from allura.model.repository import (
+        topological_sort,
+        GitLikeTree,
+        zipdir,
+        prefix_paths_union,
+    )
 
 log = logging.getLogger(__name__)
 
@@ -360,4 +367,39 @@ class HgImplementation(M.RepositoryImplementation):
             return
         self._repo.default_branch_name = name
 
+    def _get_last_commit(self, commit_id, paths):
+        # walkchangerevs appears to occasionally, randomly throw either a
+        # RevlogError or an mpatch.mpatchError complaining about corrupt
+        # repo data.  However, it's not reproducable and even just retrying
+        # the call once immediately is enough to clear the error.
+        # http://stackoverflow.com/questions/17073368/mercurial-complains-repository-is-corrupt-when-its-not
+        # seems to indicate it may be related to optimizations in the network
+        # stack, but I wasn't able to apply that solution, so this retry loop
+        # is a work-around.
+        for attempt in xrange(1, 3):
+            try:
+                ctx = cmdutil.walkchangerevs(
+                        self._hg,
+                        cmdutil.match(self._hg, pats=['path:%s' % p for p in paths]),
+                        {'rev': ['%s:0' % commit_id]},
+                        lambda ctx, fns: None,
+                    ).next()
+            except StopIteration:
+                return None, set()
+            except Exception as e:
+                # only catch these two types; unfortunately, can't seem to
+                # import mpatchError because it's raised in an .so library
+                if not (isinstance(e, RevlogError) or type(e).__name__ == 'mpatchError'):
+                    raise
+                if attempt == 2:
+                    log.exception('Error in mercurial repo: %s on %s in %s', e, commit_id, self._repo.full_fs_path)
+                    return None, set()
+                else:
+                    log.error('Potentially spurious error (retrying): %s on %s in %s', e, commit_id, self._repo.full_fs_path)
+            else:
+                return ctx.hex(), set(ctx.files())
+
+    def get_changes(self, commit_id):
+        return self._hg[commit_id].files()
+
 Mapper.compile_all()
